from __future__ import annotations

import pytest

from .utils import get_fixturenames_closure_and_arg2fixturedefs


def collect_fixtures(config: pytest.Config, items: list[pytest.Item]):
    fm = config.pluginmanager.get_plugin("funcmanage")
    func_items = [i for i in items if isinstance(i, pytest.Function)]
    for item in func_items:
        for marker in item.own_markers:
            if marker.name != "parametrize":
                continue
            params = marker.args[1] if len(marker.args) > 1 else marker.kwargs["argvalues"]
            arg2fixturedefs = {}
            for param in params:
                _, _arg2fixturedefs = get_fixturenames_closure_and_arg2fixturedefs(fm, item.parent, param)
                arg2fixturedefs.update(_arg2fixturedefs)
            item._fixtureinfo.name2fixturedefs.update(arg2fixturedefs)



from __future__ import annotations

import copy

import pytest

from .utils import get_fixturenames_closure_and_arg2fixturedefs, uniq


def normalize_metafunc_calls(metafunc, used_keys: set[str] | None = None):
    newcalls = []
    for callspec in metafunc._calls:
        calls = _normalize_call(callspec, metafunc, used_keys)
        newcalls.extend(calls)
    metafunc._calls = newcalls


def _copy_metafunc(metafunc):
    copied = copy.copy(metafunc)
    copied.fixturenames = copy.copy(metafunc.fixturenames)
    copied._calls = []
    copied._arg2fixturedefs = copy.copy(metafunc._arg2fixturedefs)
    return copied


def _normalize_call(callspec, metafunc, used_keys: set[str] | None):
    fm = metafunc.config.pluginmanager.get_plugin("funcmanage")

    used_keys = used_keys or set()
    params = callspec.params.copy() if pytest.version_tuple >= (8, 0, 0) else {**callspec.params, **callspec.funcargs}
    valtype_keys = params.keys() - used_keys

    for arg in valtype_keys:
        value = params[arg]
        fixturenames_closure, arg2fixturedefs = get_fixturenames_closure_and_arg2fixturedefs(
            fm, metafunc.definition.parent, value
        )

        if fixturenames_closure and arg2fixturedefs:
            extra_fixturenames = [fname for fname in uniq(fixturenames_closure) if fname not in params]

            newmetafunc = _copy_metafunc(metafunc)
            # Only for deadfixtures call, to not break logic
            if getattr(metafunc.config.option, "deadfixtures", False):
                newmetafunc.definition._fixtureinfo.name2fixturedefs.update(arg2fixturedefs)
            newmetafunc.fixturenames = extra_fixturenames
            newmetafunc._arg2fixturedefs.update(arg2fixturedefs)
            newmetafunc._calls = [callspec]
            fm.pytest_generate_tests(newmetafunc)
            normalize_metafunc_calls(newmetafunc, used_keys | {arg})
            return newmetafunc._calls

        used_keys.add(arg)
    return [callspec]



from __future__ import annotations

import pytest

from .fixture_collector import collect_fixtures
from .lazy_fixture import LazyFixtureWrapper
from .loader import load_lazy_fixtures
from .normalizer import normalize_metafunc_calls


@pytest.hookimpl(tryfirst=True)
def pytest_fixture_setup(fixturedef: pytest.FixtureDef, request: pytest.FixtureRequest):  # noqa: ARG001
    val = getattr(request, "param", None)
    if val is not None:
        request.param = load_lazy_fixtures(val, request)


def pytest_make_parametrize_id(config: pytest.Config, val, argname):  # noqa: ARG001
    if isinstance(val, LazyFixtureWrapper):
        return val.name
    return None


@pytest.hookimpl(hookwrapper=True)
def pytest_generate_tests(metafunc: pytest.Metafunc):
    yield

    normalize_metafunc_calls(metafunc)


def pytest_collection_modifyitems(session: pytest.Session, config: pytest.Config, items: list[pytest.Item]):  # noqa: ARG001
    if not getattr(config.option, "deadfixtures", False):
        return
    collect_fixtures(config, items)



import pytest

from .lazy_fixture import LazyFixtureWrapper
from .lazy_fixture_callable import LazyFixtureCallableWrapper


def load_lazy_fixtures(value, request: pytest.FixtureRequest):
    if isinstance(value, LazyFixtureCallableWrapper):
        return value.get_func(request)(
            *load_lazy_fixtures(value.args, request),
            **load_lazy_fixtures(value.kwargs, request),
        )
    if isinstance(value, LazyFixtureWrapper):
        return value.load_fixture(request)
    if type(value) is dict:
        for key in list(value):
            value[load_lazy_fixtures(key, request)] = load_lazy_fixtures(value.pop(key), request)
        return value
    if type(value) is list:
        for i, val in enumerate(value):
            value[i] = load_lazy_fixtures(val, request)
        return value
    if type(value) in (set, tuple):
        new_value = type(value)(load_lazy_fixtures(val, request) for val in value)
        return value if value == new_value else new_value
    return value



from __future__ import annotations

from typing import Any, Iterable, Iterator

import pytest

from .lazy_fixture import LazyFixtureWrapper
from .lazy_fixture_callable import LazyFixtureCallableWrapper


def uniq(values: Iterable[str]) -> Iterator[str]:
    seen = set()
    for value in values:
        if value not in seen:
            seen.add(value)
            yield value


def _getfixtureclosure(fm, parent_node, value_name):
    if pytest.version_tuple >= (8, 0, 0):
        fixturenames_closure, arg2fixturedefs = fm.getfixtureclosure(parent_node, [value_name], {})
    else:  # pragma: no cover
        # TODO: add tox
        _, fixturenames_closure, arg2fixturedefs = fm.getfixtureclosure([value_name], parent_node)
    return fixturenames_closure, arg2fixturedefs


def get_fixturenames_closure_and_arg2fixturedefs(fm, parent_node, value) -> tuple[list[str], dict[str, Any]]:
    if isinstance(value, LazyFixtureCallableWrapper):
        extra_fixturenames_args, arg2fixturedefs_args = get_fixturenames_closure_and_arg2fixturedefs(
            fm,
            parent_node,
            value.args,
        )
        extra_fixturenames_kwargs, arg2fixturedefs_kwargs = get_fixturenames_closure_and_arg2fixturedefs(
            fm,
            parent_node,
            value.kwargs,
        )
        extra_fixturenames_func, arg2fixturedefs_func = [], {}
        if value._func is None:
            extra_fixturenames_func, arg2fixturedefs_func = _getfixtureclosure(fm, parent_node, value.fixture_name)
        return [*extra_fixturenames_args, *extra_fixturenames_kwargs, *extra_fixturenames_func], {
            **arg2fixturedefs_args,
            **arg2fixturedefs_kwargs,
            **arg2fixturedefs_func,
        }
    if isinstance(value, LazyFixtureWrapper):
        return _getfixtureclosure(fm, parent_node, value.fixture_name)
    extra_fixturenames, arg2fixturedefs = [], {}
    # we need to check exact type
    if type(value) is dict:
        value = list(value.values()) + list(value.keys())
    # we need to check exact type
    if type(value) in {list, tuple, set}:
        for val in value:
            ef, arg2f = get_fixturenames_closure_and_arg2fixturedefs(fm, parent_node, val)
            extra_fixturenames.extend(ef)
            arg2fixturedefs.update(arg2f)
    return extra_fixturenames, arg2fixturedefs



from __future__ import annotations

from inspect import isfunction
from typing import TYPE_CHECKING, Callable

from .lazy_fixture import LazyFixtureWrapper

if TYPE_CHECKING:
    import pytest


class LazyFixtureCallableWrapper(LazyFixtureWrapper):
    _func: Callable | None
    args: tuple
    kwargs: dict

    def __init__(self, callable_or_name: Callable | str, *args, **kwargs):
        if callable(callable_or_name):
            self._func = callable_or_name
            self.name = (
                callable_or_name.__name__ if isfunction(callable_or_name) else callable_or_name.__class__.__name__
            )
        else:
            self.name = callable_or_name
            self._func = None
        self.args = args
        self.kwargs = kwargs

    def get_func(self, request: pytest.FixtureRequest) -> Callable:
        func = self._func
        if func is None:
            func = self.load_fixture(request)
            if not callable(func):
                msg = "Passed fixture is not callable"
                raise TypeError(msg)
        return func


def lfc(name: Callable | str, *args, **kwargs) -> LazyFixtureCallableWrapper:
    """lfc is a lazy fixture callable."""
    return LazyFixtureCallableWrapper(name, *args, **kwargs)



from __future__ import annotations

from dataclasses import dataclass
from operator import attrgetter
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import pytest


@dataclass
class LazyFixtureWrapper:
    name: str

    @property
    def fixture_name(self) -> str:
        return self.name.split(".", maxsplit=1)[0]

    def _get_attr(self, fixture) -> str | None:
        splitted = self.name.split(".", maxsplit=1)
        if len(splitted) == 1:
            return fixture
        return attrgetter(splitted[1])(fixture)

    def __repr__(self) -> str:
        return self.name

    def load_fixture(self, request: pytest.FixtureRequest):
        return self._get_attr(request.getfixturevalue(self.fixture_name))

    def __hash__(self) -> int:
        return hash(self.name)


def lf(name: str) -> LazyFixtureWrapper:
    """lf is a lazy fixture."""
    return LazyFixtureWrapper(name)
